@@ -16,7 +16,7 @@ use AutoLoader qw(AUTOLOAD) ;
 our @ISA = qw(Exporter) ;
 our %EXPORT_TAGS = ('all' => [ qw() ]) ;
 our @EXPORT_OK = ( @{ $EXPORT_TAGS{'all'} } ) ;
-our @EXPORT = qw(AddRule AddRuleTo AddSubpbsRule ReplaceRule ReplaceRuleTo RemoveRule BuildOk) ;
+our @EXPORT = qw(AddRule AddRuleTo AddSubpbsRule AddSubpbsRules ReplaceRule ReplaceRuleTo RemoveRule BuildOk) ;
 our $VERSION = '0.09' ;
 
 use File::Basename ;
@@ -28,7 +28,8 @@ use PBS::Shell ;
 use PBS::PBSConfig ;
 use PBS::Output ;
 use PBS::Constants ;
-use PBS::Plugin;
+use PBS::Plugin ;
+use PBS::Rules::Creator ;
 
 use base qw(PBS::Attributes) ;
 
@@ -128,7 +129,9 @@ $file_name =~ s/'$// ;
 my $class = 'User' ;
 
 my @rule_definition = @_ ;
-RunUniquePluginSub('AddRule', $file_name, $line, \@rule_definition) ;
+
+my $pbs_config = GetPbsConfig($package) ;
+RunUniquePluginSub($pbs_config , 'AddRule', $file_name, $line, \@rule_definition) ;
 
 my $first_argument = shift @rule_definition ;
 my ($name, $rule_type) ;
@@ -182,7 +185,9 @@ unless('' eq ref $class)
 	}
 
 my @rule_definition = @_ ;
-RunUniquePluginSub('AddRule', $file_name, $line, \@rule_definition) ;
+
+my $pbs_config = GetPbsConfig($package) ;
+RunUniquePluginSub($pbs_config, 'AddRule', $file_name, $line, \@rule_definition) ;
 
 my $first_argument = shift @rule_definition;
 my ($name, $rule_type) ;
@@ -230,7 +235,8 @@ $file_name =~ s/'$// ;
 my $class = 'User' ;
 
 my @rule_definition = @_ ;
-RunUniquePluginSub('AddRule', $file_name, $line, \@rule_definition) ;
+my $pbs_config = GetPbsConfig($package) ;
+RunUniquePluginSub($pbs_config , 'AddRule', $file_name, $line, \@rule_definition) ;
 
 my $first_argument = shift @rule_definition ;
 
@@ -281,7 +287,8 @@ $file_name =~ s/'$// ;
 my $class = shift ;
 
 my @rule_definition = @_ ;
-RunUniquePluginSub('AddRule', $file_name, $line, \@rule_definition) ;
+my $pbs_config = GetPbsConfig($package) ;
+RunUniquePluginSub($pbs_config, 'AddRule', $file_name, $line, \@rule_definition) ;
 
 my $first_argument = shift @rule_definition ;
 my ($name, $rule_type) ;
@@ -349,29 +356,79 @@ if(exists $package_rules{$package}{$class})
 		}
 	}
 
+my %rule_type ;
+for my $rule_type (@$rule_types)
+	{
+	$rule_type{$rule_type}++
+	}
+
+#>>>>>>>>>>>>>
+# special handling for CREATOR  rules
+# if a rule is [CREATOR] and no creator was defined in the depender definition,
+# we put a creator in the depender definition and give the builder as argument to the creator
+
+# this lets us write :
+# AddRule [CREATOR], [ 'a' =>' b'], 'touch %FILE_TO_BUILD' ;
+# and have the creator handle the digest part and call the builder to create the node
+
+if($rule_type{__CREATOR})
+	{
+	if('ARRAY' eq ref $depender_definition)
+		{
+		if('ARRAY' eq ref $depender_definition->[0])
+			{
+			die ERROR "[CREATOR] rules can't have a creator defined within depender!\n" ;
+			}
+			
+		if(defined $builder_definition)
+			{
+			#Let there be magic!
+			unshift @$depender_definition, [GenerateCreator($builder_definition)] ;
+			$builder_definition = undef ;
+			}
+		else
+			{
+			die ERROR "[CREATOR] rules must have a builder!\n" ;
+			}
+		}
+	else
+		{
+		die ERROR "[CREATOR] rules must have depender in form ['object_to_create => dependencies]!\n" ;
+		}
+	}
+#<<<<<<<<<<<<<<<<<<<<<<
+
 my ($builder_sub, $node_subs1, $builder_generated_types) = GenerateBuilder(undef, $builder_definition, $package, $name, $file_name, $line) ;
 $builder_generated_types ||= {} ;
 
-my ($depender_sub, $node_subs2, $depender_generated_types) = GenerateDepender($file_name, $line, $package, $class, $rule_types, $name, $depender_definition, $builder_sub) ;
+my ($depender_sub, $node_subs2, $depender_generated_types) = GenerateDepender($file_name, $line, $package, $class, $rule_types, $name, $depender_definition) ;
+$depender_generated_types  ||= [] ; 
 
 my $origin = '' ;
-if($pbs_config->{ADD_ORIGIN})
+$origin = ":$package:$class:$file_name:$line"  if($pbs_config->{ADD_ORIGIN}) ;
+	
+for my $rule_type (@$rule_types)
 	{
-	$origin = ":$package:$class:$file_name:$line" ;
+	$rule_type{$rule_type}++
 	}
 	
-my ($post_depend, $meta_slave, $creator) ;
-
-for my $rule_type (@$rule_types, @$depender_generated_types)
+if($rule_type{__VIRTUAL} && $rule_type{__LOCAL})
 	{
-	$creator++     if $rule_type eq CREATOR ;
-	$meta_slave++  if $rule_type eq META_SLAVE ;
-	$post_depend++ if $rule_type eq POST_DEPEND ;
+	PrintError("Rule can't be 'VIRTUAL' and 'LOCAL'.") ;
+	PbsDisplayErrorWithContext($file_name,$line) ;
+	die ;
+	}
+	
+if($rule_type{__POST_DEPEND} && $rule_type{__CREATOR})
+	{
+	PrintError("Rule can't be 'POST_DEPEND' and 'CREATOR'.") ;
+	PbsDisplayErrorWithContext($file_name,$line) ;
+	die ;
 	}
 
-if($post_depend && $creator)
+if($rule_type{__VIRTUAL} && $rule_type{__CREATOR})
 	{
-	Carp::carp ERROR("Rule can't be 'POST_DEPEND' and define a creator.") ;
+	PrintError("Rule can't be 'VIRTUAL' and 'CREATOR'.") ;
 	PbsDisplayErrorWithContext($file_name,$line) ;
 	die ;
 	}
@@ -428,18 +485,18 @@ $rule_definition->{NODE_SUBS} = $node_subs if @$node_subs ;
 if(defined $pbs_config->{DEBUG_DISPLAY_RULES})
 	{
 	my $class_info = "[$class" ;
-	$class_info .= ' (POST_DEPEND)' if $post_depend ;
-	$class_info .= ' (META_SLAVE)' if $meta_slave ;
-	$class_info .= ' (CREATOR)' if $creator ;
+	$class_info .= ' (POST_DEPEND)' if $rule_type{POST_DEPEND} ;
+	$class_info .= ' (META_SLAVE)'  if $rule_type{META_SLAVE} ;
+	$class_info .= ' (CREATOR)'     if $rule_type{CREATOR};
 	$class_info .= ']' ;
 		
 	if('HASH' eq ref $depender_definition)
 		{
-		PrintInfo("Registering subpbs rule: $class_info $name$origin")  ;
+		PrintInfo("Registering subpbs rule: $class_info '$name$origin'.")  ;
 		}
 	else
 		{
-		PrintInfo("Registering rule: $class_info $name$origin")  ;
+		PrintInfo("Registering rule: $class_info '$name$origin'.")  ;
 		}
 		
 	PrintInfo(DumpTree($rule_definition)) if defined $pbs_config->{DEBUG_DISPLAY_RULE_DEFINITION} ;
@@ -447,6 +504,8 @@ if(defined $pbs_config->{DEBUG_DISPLAY_RULES})
 	}
 
 push @{$package_rules{$package}{$class}}, $rule_definition ;
+
+return($rule_definition) ;
 }
 
 #-------------------------------------------------------------------------------
@@ -526,22 +585,40 @@ return
 
 
 #-------------------------------------------------------------------------------
+sub AddSubpbsRules
+{
+my ($package, $file_name, $line) = caller() ;
+$file_name =~ s/^'// ;
+$file_name =~ s/'$// ;
+
+for(@_)
+	{
+	__AddSubpbsRule($package, $file_name, $line, $_) ;
+	}
+}
 
 sub AddSubpbsRule
 {
+my ($package, $file_name, $line) = caller() ;
+$file_name =~ s/^'// ;
+$file_name =~ s/'$// ;
+
+__AddSubpbsRule($package, $file_name, $line, \@_) ;
+}
+
+sub __AddSubpbsRule
+{
 # Syntactic sugar, this function can be called instead for 
 # AddRule .. { subpbs_definition}
 # the compulsory arguments come first, then one can pass 
 # key-value pairs as in a normal subpbs definition
 
-my ($package, $file_name, $line) = caller() ;
-$file_name =~ s/^'// ;
-$file_name =~ s/'$// ;
+my ($package, $file_name, $line, $rule_definition) = @_ ;
 
-my $rule_definition = \@_ ;
+my $pbs_config = GetPbsConfig($package) ;
 
 my ($rule_name, $node_regex, $Pbsfile, $pbs_package, @other_setup_data) 
-	= RunUniquePluginSub('AddSubpbsRule', $file_name, $line, $rule_definition) ;
+	= RunUniquePluginSub($pbs_config, 'AddSubpbsRule', $file_name, $line, $rule_definition) ;
 
 RegisterRule
 	(
@@ -568,7 +645,7 @@ RegisterRule
 __END__
 =head1 NAME
 
-PBS::Rules::Rules - Manipulate PBS rules
+PBS::Rules - Manipulate PBS rules
 
 =head1 SYNOPSIS
 
